#include <stdio.h>
#include <string.h>
#include <getopt.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include "mikroc.h"
#include "y.tab.h"

//#define USE_LD

#define PROG_NAME "mikroc"
#define PRINT_FSTR "printfstr"
#define PRINTX_FSTR "printxfstr"
#define PRINTC_FSTR "printcfstr"
#define VAR_PREFIX "v"
#define LABEL_PREFIX "l"
#define ASMLEN 100

struct options_t {
  char compile;
  char asm_only;
  char verbose;
  char *output;
  int file_index;
};

int cur_label = 1, cur_var = 1;
FILE *asm_file;

int
yyerror(char *s)
{
  return fprintf(stderr, "%s: yyerror: %s\n", PROG_NAME, s);
}

void
cond_copy(char *dest, char *src1, char *src2)
{
  if (*src1 != 0) strncpy(dest, src1, STRLEN); 
  else strncpy(dest, src2, STRLEN);
}

int
new_label()
{
  return cur_label++;
}

void
new_var(char *var)
{
  snprintf(var, STRLEN, "[%s%i]", VAR_PREFIX, cur_var);
  cur_var++;
}

void
do_peephole(char *s)
{
  char *p, *q, tmp[10];
  int i = 0;
  
  if (((p = strstr(s, "add")) || (p = strstr(s, "sub")) || 
       (p = strstr(s, "or"))) && (p = s) && (q = strstr(s, ", 0")) &&
      (*(q + 3) == 0)) *s = 0;
  if ((p = strstr(s, "and")) && (p = s) && (q = strstr(s, ", 0")) &&
      (*(q + 3) == 0)) {
    for (p += 4; p != q; p++, i++) tmp[i] = *p; tmp[i] = 0;
    snprintf(s, ASMLEN, "xor %s, %s", tmp, tmp);
  }
  if ((p = strstr(s, "add")) && (p = s) && (q = strstr(s, ", 1")) && 
      (*(q + 3) == 0)) { 
    for (p += 4; p != q; p++, i++) tmp[i] = *p; tmp[i] = 0;
    snprintf(s, ASMLEN, "inc %s", tmp);
  }
  if ((p = strstr(s, "sub")) && (p = s) && (q = strstr(s, ", 1")) && 
      (*(q + 3) == 0)) { 
    for (p += 4; p != q; p++, i++) tmp[i] = *p; tmp[i] = 0;
    snprintf(s, ASMLEN, "dec %s", tmp);
  }
}

void
out_asm(char *s)
{
  do_peephole(s);
  fprintf(asm_file, "%s\n", s);
}

void
gen_out_start()
{
  out_asm("bits 32");
#ifdef USE_LD
  out_asm("global _start");
#else
  out_asm("global main:function");
#endif
  out_asm("extern printf");
  out_asm("");
  out_asm("section .text");
  out_asm("align 4");
  out_asm("");
#ifdef USE_LD
  out_asm("_start:");
#else
  out_asm("main:");
  out_asm("push ebp");
  out_asm("mov ebp, esp");
  out_asm("pusha");
#endif
  out_asm("");
}

void
gen_out_end()
{
  int i;
  char s[ASMLEN+1];
  
  out_asm("");
#ifdef USE_LD
  out_asm("mov ebx, 0");
  out_asm("mov eax, 1");
  out_asm("int 0x80");
#else
  out_asm("popa");
  out_asm("leave");
  out_asm("ret");
#endif
  out_asm("");
  out_asm("section .data");
  out_asm("");
  snprintf(s, ASMLEN, "%s db \"%%i\", 0", PRINT_FSTR); out_asm(s);
  snprintf(s, ASMLEN, "%s db \"%%x\", 0", PRINTX_FSTR); out_asm(s);
  snprintf(s, ASMLEN, "%s db \"%%c\", 0", PRINTC_FSTR); out_asm(s);
  out_asm("");
  out_asm("section .bss");
  out_asm("");
  for (i = 'a'; i <= 'z'; i++) { 
    snprintf(s, ASMLEN, "%c resd 1", i); out_asm(s);
  }
  for (i = 'A'; i <= 'Z'; i++) {
    snprintf(s, ASMLEN, "%c resd 1", i); out_asm(s);
  }
  for (i = 1; i < cur_var; i++) {
    snprintf(s, ASMLEN, "%s%i resd 1", VAR_PREFIX, i); out_asm(s);
  }
}

void
gen_label(int label)
{
  char s[ASMLEN+1];

  snprintf(s, ASMLEN, ".%s%i:", LABEL_PREFIX, label); out_asm(s);
}

void
gen_jmp(char *cond, int val, int label)
{
  char s[ASMLEN+1];
  int lab;

  if (cond != NULL) {
    snprintf(s, ASMLEN, "mov eax, %s", cond); out_asm(s);
    out_asm("cmp eax, dword 0");
    lab = new_label();
    if (val) snprintf(s, ASMLEN, "jz .%s%i", LABEL_PREFIX, lab);
    else snprintf(s, ASMLEN, "jnz .%s%i", LABEL_PREFIX, lab);
    out_asm(s);
    snprintf(s, ASMLEN, "jmp .%s%i", LABEL_PREFIX, label); out_asm(s);
    gen_label(lab);
  }
  else { snprintf(s, ASMLEN, "jmp .%s%i", LABEL_PREFIX, label); out_asm(s); }
}

void
gen_op(int op, char *out, char *in1, char *in2)
{
  char s[ASMLEN+1];
  int label;

  switch (op) {
  case '*':
  case '/':
  case MUL:
  case DIV:
    if ((op == '*') || (op == '/')) 
      snprintf(s, ASMLEN, "mov eax, %s", in1);
    else snprintf(s, ASMLEN, "mov eax, %s", out);
    out_asm(s);
    if ((op == '*') || (op == '/')) 
      snprintf(s, ASMLEN, "mov ebx, %s", in2);
    else snprintf(s, ASMLEN, "mov ebx, %s", in1);
    out_asm(s);
    if ((op == '*') || (op == MUL)) out_asm("imul ebx");
    else { out_asm("xor edx, edx"); out_asm("idiv ebx"); }
    snprintf(s, ASMLEN, "mov %s, eax", out); out_asm(s);
    break;
  case '+':
  case '-':
  case ADD:
  case SUB:
    if ((op == '+') || (op == '-')) 
      snprintf(s, ASMLEN, "mov eax, %s", in1);
    else snprintf(s, ASMLEN, "mov eax, %s", out);
    out_asm(s);
    if (op == '+') snprintf(s, ASMLEN, "add eax, %s", in2);
    else if (op == ADD) snprintf(s, ASMLEN, "add eax, %s", in1);
    else if (op == '-') snprintf(s, ASMLEN, "sub eax, %s", in2);
    else snprintf(s, ASMLEN, "sub eax, %s", in1);
    out_asm(s);
    snprintf(s, ASMLEN, "mov %s, eax", out); out_asm(s);
    break;
  case '<':
  case '>':
  case LE:
  case GE:
  case EQ:
  case NE:
    snprintf(s, ASMLEN, "mov eax, %s", in1); out_asm(s);
    out_asm("xor ebx, ebx");
    snprintf(s, ASMLEN, "cmp eax, dword %s", in2); out_asm(s);
    label = new_label();
    if (op == '<') snprintf(s, ASMLEN, "jnl .%s%i", LABEL_PREFIX, label);
    else if (op == '>') snprintf(s, ASMLEN, "jng .%s%i", LABEL_PREFIX, label);
    else if (op == LE) snprintf(s, ASMLEN, "jnle .%s%i", LABEL_PREFIX, label);
    else if (op == GE) snprintf(s, ASMLEN, "jnge .%s%i", LABEL_PREFIX, label);
    else if (op == EQ) snprintf(s, ASMLEN, "jne .%s%i", LABEL_PREFIX, label);
    else snprintf(s, ASMLEN, "je .%s%i", LABEL_PREFIX, label);
    out_asm(s);
    out_asm("inc ebx");
    gen_label(label);
    snprintf(s, ASMLEN, "mov %s, ebx", out); out_asm(s);
    break;
  case '&':
  case '|':
  case INV:
    snprintf(s, ASMLEN, "mov eax, %s", in1); out_asm(s);
    if (op == '&') snprintf(s, ASMLEN, "and eax, %s", in2);
    else if (op == '|') snprintf(s, ASMLEN, "or eax, %s", in2);
    else snprintf(s, ASMLEN, "neg eax");
    out_asm(s);
    snprintf(s, ASMLEN, "mov %s, eax", out); out_asm(s);
    break;
  case '=':
    snprintf(s, ASMLEN, "mov eax, %s", in1); out_asm(s);
    snprintf(s, ASMLEN, "mov %s, eax", out); out_asm(s);
    break;
  case PRINT:
  case PRINTX:
  case PRINTC:
    snprintf(s, ASMLEN, "push dword %s", in1); out_asm(s);
    if (op == PRINT) snprintf(s, ASMLEN, "push dword %s", PRINT_FSTR);
    else if (op == PRINTX) snprintf(s, ASMLEN, "push dword %s", PRINTX_FSTR);
    else snprintf(s, ASMLEN, "push dword %s", PRINTC_FSTR);
    out_asm(s);
    out_asm("call printf");
    out_asm("add esp, 8");
    break;
  default:
    fprintf(stderr, "%s: unknown operation `%i'\n", PROG_NAME, op);
  }
}

void
print_help()
{
  printf("Usage: %s [options] file...\n", PROG_NAME);
  printf("Options:\n");
  printf("  -h  --help          Display this information\n");
  printf("  -c  --compile       Compile and assemble, but do not link\n");
  printf("  -S  --asm-only      Compile only; do not assemble or link\n");
  printf("  -v  --verbose       Display the programs invoked by the compiler and other informations\n");
  printf("  -o <file>  --output Place the output into <file> (applies only to the last given source file)\n");
  printf("\nFor additional informations, please see README file.\n");
}

void
parse_options(struct options_t *options, int argc, char *argv[])
{
  int o;
  struct option long_options[] = {
    {"help", no_argument, NULL, 'h'},
    {"compile", no_argument, NULL, 'c'},
    {"asm-only", no_argument, NULL, 'S'},
    {"verbose", no_argument, NULL, 'v'},
    {"output", required_argument, NULL, 'o'},
    {0, 0, 0, 0}
  };

  options->compile = 0; options->asm_only = 0; options->verbose = 0; 
  options->output = NULL; options->file_index = 0;
  opterr = 0;
  while ((o = getopt_long(argc, argv, "hcSvo:", long_options, NULL)) != -1) {
    switch (o) {
    case 'h': print_help(); exit(0); break;
    case 'c': options->compile = 1; break;
    case 'S': options->asm_only = 1; break;
    case 'v': options->verbose = 1; break;
    case 'o': options->output = optarg; break;
    case '?': 
      fprintf(stderr, "%s: argument to `-%c' is missing\n", PROG_NAME, 
	      optopt); 
      break;
    default: ;
    }
  }
  if (optind < argc) options->file_index = optind;
}

char *
replace_suffix(char *dest, char *src, char *old, char *new)
{
  char *dot, *end;
  char *suf;

  strncpy(dest, src, 256);
  dot = dest; while (*dot) dot++; suf = old; while (*suf) suf++;
  end = dot;
  while ((suf != old) && (*dot == *suf)) { dot--; suf--; }
  if (suf != old) dot = end;
  while (*new) *dot++ = *new++; *dot = 0;
  return dest;
}

int
main(int argc, char *argv[])
{
  struct options_t options;
  char *src_fn;
  char asm_fn[256] = "a.out.s";
  char obj_fn[256];
  char elf_fn[256];
  char *nasm_args[5];
#ifdef USE_LD
  char *ld_args[8];
#else
  char *gcc_args[5];
#endif
  pid_t child_pid;
  int child_status;
  
  parse_options(&options, argc, argv);
  options.file_index--;
  if (options.verbose) printf("%s version 0.1.0\n", PROG_NAME);
  while (options.file_index + 1 < argc) {
    options.file_index++;
    if (!options.file_index) {
      if (!options.verbose) {
	fprintf(stderr, "%s: no input files\n", PROG_NAME); exit(1);
      }
      else exit(0);
    }
    src_fn = argv[options.file_index];
    if (strcmp(src_fn, "-")) {
      if (freopen(src_fn, "r", stdin) == NULL) { 
	fprintf(stderr, "%s: freopen error (%s)\n", PROG_NAME, src_fn); 
	continue;
      }
      replace_suffix(asm_fn, src_fn, ".mc", ".s");
    }
    if ((asm_file = fopen(asm_fn, "w")) == NULL) {
      fprintf(stderr, "%s: fopen error (%s)\n", PROG_NAME, asm_fn); continue;
    }
    if (options.verbose) printf("generating assembler file %s\n", asm_fn);
    child_status = yyparse(); 
    fclose(asm_file);
    if (child_status) continue;
    if (!options.asm_only) {
      replace_suffix(obj_fn, asm_fn, ".s", ".o");
      nasm_args[0] = "nasm"; nasm_args[1] = "-f"; nasm_args[2] = "elf";
      nasm_args[3] = strdup(asm_fn); nasm_args[4] = NULL;
      if (options.verbose) { 
	int i = 0; 
	for (; i < 4; i++) printf("%s ", nasm_args[i]); printf("\n");
      }
      if ((child_pid = fork()) == 0) {
	execvp("nasm", nasm_args);
	fprintf(stderr, "%s: execvp error (nasm)\n", PROG_NAME); abort();
      }
      else if (child_pid != -1) {
	waitpid(child_pid, &child_status, 0);
	if (!WIFEXITED(child_status)) { 
	  fprintf(stderr, "%s: nasm error\n", PROG_NAME); exit(1);
	}
      }
      else { fprintf(stderr, "%s: fork error\n", PROG_NAME); exit(1); }
      remove(asm_fn);
    }
    if ((!options.asm_only) && (!options.compile)) {
      if ((options.file_index == argc - 1) && (options.output != NULL)) 
	strncpy(elf_fn, options.output, 256);
      else replace_suffix(elf_fn, obj_fn, ".o", "");
#ifdef USE_LD
      ld_args[0] = "ld"; ld_args[1] = "-dynamic-linker";
      ld_args[2] = "/lib/ld-linux.so.2"; // TODO: may vary
      ld_args[3] = "-lc"; ld_args[4] = "-o"; ld_args[5] = strdup(elf_fn);
      ld_args[6] = strdup(obj_fn); ld_args[7] = NULL;
#else
      gcc_args[0] = "gcc"; gcc_args[1] = "-o"; gcc_args[2] = strdup(elf_fn);
      gcc_args[3] = strdup(obj_fn); gcc_args[4] = NULL;
#endif
      if (options.verbose) { 
	int i = 0;
#ifdef USE_LD
	for (; i < 7; i++) 
	  printf("%s ", ld_args[i]);
#else
	for (; i < 4; i++) 
	  printf("%s ", gcc_args[i]);
#endif
	printf("\n");
      }
      replace_suffix(elf_fn, obj_fn, ".o", "");
      if ((child_pid = fork()) == 0) {
#ifdef USE_LD
	execvp("ld", ld_args);
	fprintf(stderr, "%s: execvp error (ld)\n", PROG_NAME);
#else
	execvp("gcc", gcc_args);
	fprintf(stderr, "%s: execvp error (gcc)\n", PROG_NAME);
#endif
	abort();
      }
      else if (child_pid != -1) {
	waitpid(child_pid, &child_status, 0);
	if (!WIFEXITED(child_status)) { 
#ifdef USE_LD
	  fprintf(stderr, "%s: ld error\n", PROG_NAME);
#else
	  fprintf(stderr, "%s: gcc error\n", PROG_NAME);
#endif
	  exit(1);
	}
      }
      else { fprintf(stderr, "%s: fork error\n", PROG_NAME); exit(1); }
      remove(obj_fn);
    }
  }

  return 0;
}
